# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Federated algorithm interface."""

from typing import Any, Callable, Iterable, Mapping, Tuple

from fedjax import core

PyTree = Any
# AlgorithmState contains any round specific parameters (e.g. model parameters)
# that will be passed from round to round. This state must be serializable for
# checkpointing. We strongly recommend making this a PyTree.
AlgorithmState = PyTree
# Client contains the client identifier and dataset as well as any per client
# information like persistent client state for the cross-silo setting.
Client = Tuple[str, Any]
# ClientDiagnostics contains the accumulated per step results keyed by client
# identifiers.
ClientDiagnostics = Mapping[str, Any]


@core.dataclass
class FederatedAlgorithm:
  """Container for all federated algorithms.

  `FederatedAlgorithm` defines the required methods that need to be implemented
  by all custom federated algorithms. Defining federated algorithms in this
  structure will allow implementations to work seamlessly with the convenience
  methods in the `fedjax.training` API, like checkpointing.

  Example toy implementation:

  ```python
  # Federated algorithm that just counts total number of points across clients
  # across rounds.

  def count_federated_algorithm():

    def init(init_count):
      return {'count': init_count}

    def run_one_round(state, clients):
      count = 0
      client_diagnostics = {}
      # Count sizes across clients.
      for client_id, client_dataset in clients:
        client_count = 0
        for d in client_dataset:
          client_count += sum(l.size for l in jax.tree_util.tree_leaves(d))
        # Summation across clients in one round.
        count += client_count
        client_diagnostics[client_id] = client_count
      # Summation across rounds.
      state = {'count': state['count'] + count}
      return state, client_diagnostics

    return FederatedAlgorithm(init, run_one_round)

  all_clients = [
      [
          ('cid0', [jnp.array([1, 2]), jnp.array([1, 2, 3, 4])]),
          ('cid1', [jnp.array([1, 2, 3, 4, 5])]),
          ('cid2', [jnp.array([1]), jnp.array([1, 2])]),
      ],
      [
          ('cid3', [jnp.array([1, 2, 3, 4])]),
          ('cid4', [jnp.array([1]), jnp.array([1, 2]), jnp.array([1, 2, 3])]),
          ('cid5', [jnp.array([1, 2, 3, 4, 5, 6, 7])]),
      ],
  ]
  algorithm = count_federated_algorithm()
  state = algorithm.init(0)
  for round_num in range(2):
    state, client_diagnostics = algorithm.run_one_round(state,
                                                        all_clients[round_num])
    print(round_num, state)
    print(round_num, client_diagnostics)
  # 0 {'count': 14}
  # 0 {'cid0': 6, 'cid1': 5, 'cid2': 3}
  # 1 {'count': 31}
  # 1 {'cid3': 4, 'cid4': 6, 'cid5': 7}
  ```

  Attributes:
    init: Initializes the `AlgorithmState`. Typically, the input to this
      method will be the initial model `Params`. This should only be run once
      at the beginning of training.
    run_one_round: Completes one round of federated training given an input
      `AlgorithmState` and a sequence of client identifiers to client datasets.
      The output will be a new, updated `AlgorithmState` along with any per
      client `DiagnosticInfo` (e.g. train metrics).
  """
  init: Callable[..., AlgorithmState]
  run_one_round: Callable[[AlgorithmState, Iterable[Client]],
                          Tuple[AlgorithmState, ClientDiagnostics]]



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.algorithms.agnostic_fed_avg."""

from absl.testing import parameterized
from fedjax import core
from fedjax.algorithms import agnostic_fed_avg
from fedjax.datasets import toy_regression as toy_regression_data
from fedjax.models import toy_regression as toy_regression_model
import jax.numpy as jnp
import tensorflow as tf


class AgnosticFedAvgTest(tf.test.TestCase, parameterized.TestCase):

  @parameterized.named_parameters(
      {
          'testcase_name':
              'EG',
          'domain_algorithm':
              agnostic_fed_avg.DomainAlgorithm.EG,
          'expected_domain_weights':
              [0.08078121, 0.26637138, 0.10979304, 0.54305437]
      }, {
          'testcase_name': 'NONE',
          'domain_algorithm': agnostic_fed_avg.DomainAlgorithm.NONE,
          'expected_domain_weights': [0.2, 0.4, 0.1, 0.3]
      })
  def test_update_domain_weights(self, domain_algorithm,
                                 expected_domain_weights):
    domain_weights = agnostic_fed_avg._update_domain_weights(
        domain_weights=jnp.array([0.2, 0.4, 0.1, 0.3]),
        domain_loss=jnp.array([1., 2., 3., 4.]),
        domain_learning_rate=0.5,
        domain_algorithm=domain_algorithm)

    self.assertAllClose(domain_weights, expected_domain_weights)

  def test_run(self):
    """Tests agnostic federated averaging on minmax regression.

    fedjax.datasets.toy_regression has the formal description.
    To summarize:
      1. Random points are generated along a line and assigned to 2 domains.
      2. The points in the two domains are offset set to have the same average
        value magnitude but opposite signs.
        (e.g. mean(domain_1) == -1.0 and mean(domain_2) == 1.0)
      3. Training task is to find the point value that minimizes the maximum
        distance from the domain centers. Given (2), this value should be 0.
    """
    federated_data, _ = toy_regression_data.load_data(
        num_clients=2, num_domains=2, num_points=400, seed=0)
    model = toy_regression_model.create_regression_model()
    rng_seq = core.PRNGSequence(0)
    algorithm = agnostic_fed_avg.AgnosticFedAvg(
        federated_data=federated_data,
        model=model,
        client_optimizer=core.get_optimizer(
            core.OptimizerName.SGD, learning_rate=0.3),
        # NB: Momentum doesn't play well with this toy example.
        server_optimizer=core.get_optimizer(
            core.OptimizerName.SGD, learning_rate=1.0),
        hparams=agnostic_fed_avg.AgnosticFedAvgHParams(
            # Batch size should be large enough to fit all data in one batch.
            # This is because toy_regression loss isn't an average over examples
            # meaning the multi batch evaluation logic will result in incorrect
            # "total loss" values.
            train_data_hparams=core.ClientDataHParams(batch_size=1000),
            init_domain_weights=(0.4, 0.6),
            domain_id_fn=toy_regression_data.domain_id_fn,
            domain_learning_rate=0.3,
            domain_algorithm=agnostic_fed_avg.DomainAlgorithm.EG,
        ),
        rng_seq=rng_seq,
    )

    state = algorithm.init_state()
    for _ in range(100):
      state = algorithm.run_round(state, federated_data.client_ids)

    # 0 should be the center between the max and min point averages.
    self.assertAlmostEqual(state.params['linear']['w'], 0., places=6)
    # Domain weights should even out.
    self.assertAllClose(state.domain_weights, [0.5, 0.5])


if __name__ == '__main__':
  tf.test.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for prefetching from tff.simulation.ClientData."""

import collections
from concurrent import futures
import functools
from typing import Callable, Iterable, Optional

from fedjax.core import tree_util
from fedjax.core import typing
import tensorflow as tf

DEFAULT_ASYNC_TF_DATASET_ITERATOR_PREFETCH = 2
DEFAULT_PREFETCH_THREAD_POOL_SIZE = 10
DEFAULT_PREFETCH_NUM_INIT_FETCH = 10


class AsyncTFDatasetIterator:
  """Asynchronously create a TF dataset iterator, and fetches the first n values.

  TensorFlow dataset iterators are expensive to create, and first values out of
  the iteration often take more time than the rest. We thus wrap these
  computation into a future, so that we can submit it well before we need the
  data.

  While FutureIterator can generalize to other types of iterables, they are
  useless for pure Python iterators because of GIL. Thus we only handle
  TensorFlow datasets.
  """

  def __init__(self,
               executor: futures.ThreadPoolExecutor,
               dataset_fn: Callable[[], tf.data.Dataset],
               num_prefetch: Optional[int] = None):
    """Initializes the iterator.

    Args:
      executor: Executor to submit future computation to.
      dataset_fn: Callable that creates the underlying TensorFlow dataset.
      num_prefetch: Number of values to fetch during the future call. Defaults
        to DEFAULT_ASYNC_TF_DATASET_ITERATOR_PREFETCH.
    """
    if num_prefetch is None:
      num_prefetch = DEFAULT_ASYNC_TF_DATASET_ITERATOR_PREFETCH

    def prefetch():
      """Creates the iterator and fetches the first few values."""
      iterator = dataset_fn().as_numpy_iterator()
      values = collections.deque()
      for _ in range(num_prefetch):
        try:
          v = next(iterator)
        except StopIteration:
          break
        values.append(v)
      return values, iterator

    self._pending = executor.submit(prefetch)
    self._values, self._iterator = None, None

  def __iter__(self):
    return self

  def __next__(self):
    if self._pending is not None:
      # Calling result() will also raise any uncaught exceptions in prefetch().
      # Therefore we don't need any special handling like we do for
      # PrefetchClientDatasetsIterator.
      self._values, self._iterator = self._pending.result()
      self._pending = None
    if self._values:
      return self._values.popleft()
    return next(self._iterator)


class PrefetchClientDatasetsIterator:
  """Prefetches per-client datasets from one or more fedjax.FederatedData.

  This class uses asynchrounous prefetching to hide per-client dataset
  creation/iteration overhead. At each `next()` call,
  PrefetchClientDatasetsIterator produces one iterator for each
  fedjax.FederatedData in the PyTree given during `__init__()`. These iterators
  can be used to iterate over the corresponding client dataset.

  Example usage:
    federated_data = (train_data, (eval_train_data, eval_test_data))
    client_ids = ["0", "1", ...]
    data_iterator = PrefetchClientDatasetsIterator(federated_data, client_ids)
    # Iterators for client_id "0"
    client_id_0, iters_0 = next(data_iterator)
    train_iter_0, (eval_train_iter_0, eval_test_iter_0) = iters_0
    # Iterators for client_id "1"
    client_id_1, iters_1 = next(data_iterator)
    train_iter_1, (eval_train_iter_1, eval_test_iter_1) = iters_1
  """

  def __init__(self,
               federated_data: typing.PyTree,
               client_ids: Iterable[str],
               num_threads: Optional[int] = None,
               num_init_fetch: Optional[int] = None):
    """Initializes a PrefetchClientDatasetsIterator.

    Args:
      federated_data: PyTree of fedjax.FederatedData.
      client_ids: Client ids in the order the caller expects to access.
      num_threads: Number of prefetching threads. Defaults to
        DEFAULT_PREFETCH_THREAD_POOL_SIZE.
      num_init_fetch: Number of initial fetches to schedule during __init__().
        Defaults to DEFAULT_PREFETCH_NUM_INIT_FETCH.
    """
    if num_threads is None:
      num_threads = DEFAULT_PREFETCH_THREAD_POOL_SIZE
    if num_init_fetch is None:
      num_init_fetch = DEFAULT_PREFETCH_NUM_INIT_FETCH
    self._federated_data = federated_data
    self._client_ids = iter(client_ids)
    self._executor = futures.ThreadPoolExecutor(num_threads)
    # Each value is a (value, exception) tuple.
    self._buf = collections.deque()
    self._done = False
    for _ in range(num_init_fetch):
      self._fetch()

  def __del__(self):
    self._executor.shutdown(wait=False)

  def __iter__(self):
    return self

  def __next__(self):
    v, e = self._buf.popleft()
    if e is not None:
      raise e
    self._fetch()
    return v

  def _fetch(self):
    """Fetches the iterators for the next client."""
    if self._done:
      return
    try:
      client_id = next(self._client_ids)
    except StopIteration as e:
      self._buf.append((None, e))
      self._done = True
      return

    def create_iterator(one_federated_data):
      return AsyncTFDatasetIterator(
          self._executor,
          functools.partial(one_federated_data.create_tf_dataset_for_client,
                            client_id))

    try:
      v = (client_id, tree_util.tree_map(create_iterator, self._federated_data))
      self._buf.append((v, None))
    except Exception as e:  # pylint: disable=broad-except
      self._buf.append((None, e))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface definitions for aggregators."""

import abc
from typing import Iterable, Generic, NamedTuple, Tuple, TypeVar
from fedjax import core

T = TypeVar('T')
W = TypeVar('W')


class Aggregator(Generic[T], metaclass=abc.ABCMeta):
  """Interface for algorithms to aggregate.

  This interface structures aggregator algorithms that is used each round.
  Aggregate state contains any round specific parameters (e.g. number of bits)
  that will be passed from round to round. This state is initialized by
  `init_state` and passed as input into and returned as output from `aggregate`.
  We suggest defining state using `typing.NamedTuple` as this provides
  immutability, type hinting, and works by default with JAX transformations.

  Anything else that is static and doesn't update at each round can be defined
  in the `__init__`.

  The expected usage of Aggregator is as follows:

  ```
  aggregator = Aggregator()
  state = aggregator.init_state()
  rng_seq = core.PRNGSequence(42)
  for i in range(num_rounds):
    params_and_weights = compute_client_outputs(i)
    state, aggregated_params = aggregator.aggregate(state,
                                                    params_and_weights,
                                                    rng_seq)
  ```
  """

  @abc.abstractmethod
  def init_state(self) -> T:
    """Returns initial state of aggregator."""

  @abc.abstractmethod
  def aggregate(self, aggregator_state: T,
                params_and_weight: Iterable[Tuple[W, float]],
                rng_seq: core.PRNGSequence) -> Tuple[T, W]:
    """Returns the aggregated params and new state."""


class MeanAggregatorState(NamedTuple):
  """State of default aggregator passed between rounds.

  Attributes:
    total_weight: the number of samples
  """
  total_weight: float


class MeanAggregator(Aggregator):
  """Default aggregator that sums params according to weights."""

  def init_state(self):
    return MeanAggregatorState(total_weight=0.0)

  def aggregate(
      self, aggregator_state: MeanAggregatorState,
      params_and_weight: Iterable[Tuple[W, float]],
      rng_seq: core.PRNGSequence
  ) -> Tuple[MeanAggregatorState, W]:
    """Returns (weighted) mean of input trees and weights.

    Args:
      aggregator_state: State of the input.
      params_and_weight: Iterable of tuples of pytrees and associated weights.
      rng_seq: Random sequence.

    Returns:
      New state, (Weighted) mean of input trees and weights.
    """
    del rng_seq
    # TODO(theertha): avoid the need to copying the entire sequence to memory.
    params_and_weight = list(params_and_weight)
    weights = [weight for params, weight in params_and_weight]
    new_state = MeanAggregatorState(aggregator_state.total_weight +
                                    sum(weights))
    return core.tree_mean(params_and_weight), new_state



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.evaluation_util."""

import collections

from absl.testing import parameterized
from fedjax.core import dataset_util
from fedjax.core import evaluation_util
from fedjax.core import metrics
from fedjax.core import test_util
import haiku as hk
import tensorflow as tf


class EvaluationUtilTest(tf.test.TestCase, parameterized.TestCase):

  def test_evaluate_single_client(self):
    num_clients = 10
    num_classes = 10
    num_examples = 100
    client_data_hparams = dataset_util.ClientDataHParams(
        batch_size=20, num_epochs=3)
    data, model = test_util.create_toy_example(
        num_clients=num_clients,
        num_clusters=4,
        num_classes=num_classes,
        num_examples=num_examples,
        seed=0)
    rng_seq = hk.PRNGSequence(0)
    init_params = model.init_params(next(rng_seq))
    dataset = dataset_util.preprocess_tf_dataset(
        dataset_util.create_tf_dataset_for_clients(data), client_data_hparams)

    with self.subTest('tf dataset'):
      init_metrics = evaluation_util.evaluate_single_client(
          dataset=dataset, model=model, params=init_params)
      self.assertLess(0.0, init_metrics['loss'])

    with self.subTest('plain iterator'):
      init_metrics = evaluation_util.evaluate_single_client(
          dataset=dataset.as_numpy_iterator(), model=model, params=init_params)
      self.assertLess(0.0, init_metrics['loss'])

  def test_aggregate_metrics(self):
    metrics_dict = [
        collections.OrderedDict(
            loss=metrics.MeanMetric(total=2, count=11),
            num_examples=metrics.CountMetric(count=3)),
        collections.OrderedDict(
            loss=metrics.MeanMetric(total=6, count=9),
            num_examples=metrics.CountMetric(count=7))
    ]

    aggregated = evaluation_util.aggregate_metrics(metrics_dict)

    self.assertAlmostEqual(aggregated['loss'], 0.4)
    self.assertEqual(aggregated['num_examples'], 10.)

  def test_aggregate_metrics_empty(self):
    aggregated = evaluation_util.aggregate_metrics([])

    self.assertEmpty(aggregated)


if __name__ == '__main__':
  tf.test.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Lightweight wrappers for encapsulating JAX optimizers."""

import enum
from typing import Callable, Tuple

import dataclasses
from fedjax.core.typing import OptState
from fedjax.core.typing import Params
from fedjax.core.typing import Updates
import optax


@dataclasses.dataclass(frozen=True)
class Optimizer:
  """General optimizer interface.

  This is a simple container for the init_fn and update_fn functions returned by
  optax. Use apply_updates to apply the updates to the parameters.

  Attributes:
    init_fn: Initializes (possibly empty) sets of statistics (aka state).
    update_fn: Transforms a parameter update or gradient and updates the state.
    apply_updates: Apply the transformed gradients update to the parameters.
  """

  init_fn: Callable[[Params], OptState]
  update_fn: Callable[[Updates, OptState], Tuple[Updates, OptState]]
  apply_updates: Callable[[Params, Updates], Params] = optax.apply_updates


@enum.unique
class OptimizerName(enum.Enum):
  """Supported optimizers in fedjax."""
  # Standard SGD algorithm.
  SGD = 'SGD'
  # SGD algorithm with momentum.
  MOMENTUM = 'MOMENTUM'
  # Algorithm from arxiv.org/pdf/1412.6980.pdf
  ADAM = 'ADAM'
  # Algorithm from cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf
  RMSPROP = 'RMSPROP'
  # Algorithm from https://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf
  ADAGRAD = 'ADAGRAD'


def get_optimizer(optimizer_name: OptimizerName,
                  learning_rate: float,
                  momentum: float = 0.0,
                  adam_beta1: float = 0.9,
                  adam_beta2: float = 0.999,
                  adam_epsilon: float = 1e-8,
                  rmsprop_decay: float = 0.9,
                  rmsprop_epsilon: float = 1e-8,
                  adagrad_init_accumulator: float = 0.1,
                  adagrad_epsilon: float = 1e-6) -> Optimizer:
  """Given parameters, returns the corresponding optimizer.

  Args:
    optimizer_name: One of SGD, MOMENTUM, ADAM, RMSPROP.
    learning_rate: Learning rate for all optimizers.
    momentum: Momentum parameter for MOMENTUM.
    adam_beta1: beta1 parameter for ADAM.
    adam_beta2: beta2 parameter for ADAM.
    adam_epsilon: epsilon parameter for ADAM.
    rmsprop_decay: decay parameter for RMSPROP.
    rmsprop_epsilon: epsilon parameter for RMSPROP.
    adagrad_init_accumulator: initial accumulator for ADAGRAD.
    adagrad_epsilon: epsilon parameter for ADAGRAD.

  Returns:
    Returns the Optimizer with the specified properties.

  Raises:
    ValueError: iff the optimizer names is not one of SGD, MOMENTUM, ADAM,
  RMSPROP, or Adagrad, raises errors.
  """
  if optimizer_name == OptimizerName.SGD:
    return Optimizer(*optax.sgd(learning_rate))
  elif optimizer_name == OptimizerName.MOMENTUM:
    return Optimizer(*optax.sgd(learning_rate, momentum))
  elif optimizer_name == OptimizerName.ADAM:
    return Optimizer(*optax.adam(
        learning_rate, b1=adam_beta1, b2=adam_beta2, eps=adam_epsilon))
  elif optimizer_name == OptimizerName.RMSPROP:
    return Optimizer(
        *optax.rmsprop(learning_rate, decay=rmsprop_decay, eps=rmsprop_epsilon))
  elif optimizer_name == OptimizerName.ADAGRAD:
    return Optimizer(*optax.adagrad(
        learning_rate,
        initial_accumulator_value=adagrad_init_accumulator,
        eps=adagrad_epsilon))
  else:
    raise ValueError(f'Unsupported optimizer_name {optimizer_name}.')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.algorithm.fed_avg."""

from fedjax import core
from fedjax.algorithms import fed_avg
import tensorflow as tf


class FedAvgTest(tf.test.TestCase):

  def test_run(self):
    data, model = core.test_util.create_toy_example(
        num_clients=10, num_clusters=4, num_classes=10, num_examples=5, seed=0)
    dataset = core.create_tf_dataset_for_clients(data).batch(50)
    algorithm = fed_avg.FedAvg(
        federated_data=data,
        model=model,
        client_optimizer=core.get_optimizer(
            core.OptimizerName.SGD, learning_rate=0.1),
        server_optimizer=core.get_optimizer(
            core.OptimizerName.MOMENTUM, learning_rate=2.0, momentum=0.9),
        hparams=fed_avg.FedAvgHParams(
            train_data_hparams=core.ClientDataHParams(
                batch_size=100, num_epochs=1)),
        rng_seq=core.PRNGSequence(0))

    state = algorithm.init_state()
    init_metrics = core.evaluate_single_client(
        dataset=dataset, model=model, params=state.params)
    for _ in range(10):
      state = algorithm.run_round(state, data.client_ids)

    metrics = core.evaluate_single_client(
        dataset=dataset, model=model, params=state.params)
    self.assertLess(metrics['loss'], init_metrics['loss'])
    self.assertGreater(metrics['accuracy'], init_metrics['accuracy'])


if __name__ == '__main__':
  tf.test.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""FederatedData backed by SQLite."""

import random
from typing import Callable, Iterable, Iterator, Optional, Tuple
import zlib

from fedjax.experimental import client_datasets
from fedjax.experimental import federated_data
from fedjax.experimental import serialization
import sqlite3


def decompress_and_deserialize(data: bytes):
  data = zlib.decompress(data)
  return serialization.msgpack_deserialize(data)


class SQLiteFederatedData(federated_data.FederatedData):
  """Federated dataset backed by SQLite.

  The SQLite database should contain a table named "federated_data" created with
  the following command:
  ```
  CREATE TABLE federated_data (
    client_id BLOB NOT NULL PRIMARY KEY,
    data BLOB NOT NULL,
    num_examples INTEGER NOT NULL
  );
  ```
  where,
  - `client_id` is the bytes client id.
  - `data` is the serialized client dataset examples.
  - `num_examples` is the number of examples in the client dataset.

  By default we use zlib compressed msgpack blobs for `data` (see
  decompress_and_deserialize()).
  """

  @staticmethod
  def new(
      path: str,
      parse_examples: Callable[
          [bytes], client_datasets.Examples] = decompress_and_deserialize
  ) -> 'SQLiteFederatedData':
    """Opens a federated dataset stored as an SQLite3 database.

    Args:
      path: Path to the SQLite database file.
      parse_examples: Function for deserializing client dataset examples.

    Returns:
      SQLite3DataSource.
    """
    connection = sqlite3.connect(path)
    return SQLiteFederatedData(connection, parse_examples)

  def __init__(self,
               connection: sqlite3.Connection,
               parse_examples: Callable[[bytes], client_datasets.Examples],
               start: Optional[federated_data.ClientId] = None,
               stop: Optional[federated_data.ClientId] = None,
               preprocess_client: client_datasets.Preprocessor = client_datasets
               .NoOpPreprocessor,
               preprocess_batch: client_datasets.Preprocessor = client_datasets
               .NoOpPreprocessor):
    self._connection = connection
    self._parse_examples = parse_examples
    self._start = start
    self._stop = stop
    self._preprocess_client = preprocess_client
    self._preprocess_batch = preprocess_batch

  def slice(
      self,
      start: Optional[federated_data.ClientId] = None,
      stop: Optional[federated_data.ClientId] = None) -> 'SQLiteFederatedData':
    return SQLiteFederatedData(self._connection, self._parse_examples, start,
                               stop, self._preprocess_client,
                               self._preprocess_batch)

  def preprocess_client(
      self, fn: Callable[[client_datasets.Examples], client_datasets.Examples]
  ) -> 'SQLiteFederatedData':
    return SQLiteFederatedData(self._connection, self._parse_examples,
                               self._start, self._stop,
                               self._preprocess_client.append(fn),
                               self._preprocess_batch)

  def preprocess_batch(
      self, fn: Callable[[client_datasets.Examples], client_datasets.Examples]
  ) -> 'SQLiteFederatedData':
    return SQLiteFederatedData(self._connection, self._parse_examples,
                               self._start, self._stop, self._preprocess_client,
                               self._preprocess_batch.append(fn))

  def _range_where(self) -> str:
    """Builds appropriate WHERE clauses for start/stop ranges."""
    if self._start is None and self._stop is None:
      return '(1)'
    elif self._start is not None and self._stop is not None:
      return '(:start <= client_id AND client_id < :stop)'
    elif self._start is None:
      return '(client_id < :stop)'
    else:
      return '(:start <= client_id)'

  def num_clients(self) -> int:
    cursor = self._connection.execute(
        f'SELECT COUNT(*) FROM federated_data WHERE {self._range_where()};', {
            'start': self._start,
            'stop': self._stop
        })
    return cursor.fetchone()[0]

  def client_ids(self) -> Iterator[federated_data.ClientId]:
    cursor = self._connection.execute(
        f'SELECT client_id FROM federated_data WHERE {self._range_where()} ORDER BY rowid;',
        {
            'start': self._start,
            'stop': self._stop
        })
    while True:
      result = cursor.fetchone()
      if result is None:
        break
      yield result[0]

  def client_sizes(self) -> Iterator[Tuple[federated_data.ClientId, int]]:
    cursor = self._connection.execute(
        f'SELECT client_id, num_examples FROM federated_data WHERE {self._range_where()} ORDER BY rowid;',
        {
            'start': self._start,
            'stop': self._stop
        })
    while True:
      result = cursor.fetchone()
      if result is None:
        break
      yield tuple(result)

  def client_size(self, client_id: federated_data.ClientId) -> int:
    if ((self._start is None or self._start <= client_id) and
        (self._stop is None or client_id < self._stop)):
      cursor = self._connection.execute(
          'SELECT num_examples FROM federated_data WHERE client_id = ?',
          [client_id])
      result = cursor.fetchone()
      if result is not None:
        return result[0]
    raise KeyError

  def clients(
      self
  ) -> Iterator[Tuple[federated_data.ClientId, client_datasets.ClientDataset]]:
    for k, v in self._read_clients():
      yield k, self._client_dataset(v)

  def _read_clients(self):
    cursor = self._connection.execute(
        f'SELECT client_id, data FROM federated_data WHERE {self._range_where()} ORDER BY rowid;',
        {
            'start': self._start,
            'stop': self._stop
        })
    while True:
      result = cursor.fetchone()
      if result is None:
        break
      yield tuple(result)

  def shuffled_clients(
      self,
      buffer_size: int,
      seed: Optional[int] = None
  ) -> Iterator[Tuple[federated_data.ClientId, client_datasets.ClientDataset]]:
    rng = random.Random(seed)
    while True:
      for k, v in federated_data.buffered_shuffle(self._read_clients(),
                                                  buffer_size, rng):
        yield k, self._client_dataset(v)

  def get_clients(
      self, client_ids: Iterable[federated_data.ClientId]
  ) -> Iterator[Tuple[federated_data.ClientId, client_datasets.ClientDataset]]:
    for client_id in client_ids:
      yield client_id, self.get_client(client_id)

  def get_client(
      self,
      client_id: federated_data.ClientId) -> client_datasets.ClientDataset:
    if ((self._start is None or self._start <= client_id) and
        (self._stop is None or client_id < self._stop)):
      cursor = self._connection.execute(
          'SELECT data FROM federated_data WHERE client_id = ?', [client_id])
      result = cursor.fetchone()
      if result is not None:
        return self._client_dataset(result[0])
    raise KeyError

  def _client_dataset(self, v: bytes) -> client_datasets.ClientDataset:
    examples = self._preprocess_client(self._parse_examples(v))
    return client_datasets.ClientDataset(examples, self._preprocess_batch)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.datasets.emnist."""

from fedjax.datasets import emnist
import tensorflow as tf


class EmnistDataTest(tf.test.TestCase):

  def test_flip_and_expand(self):
    dataset = tf.data.Dataset.from_tensor_slices({
        'pixels': [[[1.0, 0.9], [0.8, 1.0]]],
        'label': [2]
    })

    output = next(emnist.flip_and_expand(dataset).as_numpy_iterator())

    self.assertAllClose(output['x'], [[[0.], [0.1]], [[0.2], [0.]]])
    self.assertAllEqual(output['y'], 2)


if __name__ == '__main__':
  tf.test.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.training.logging."""

import tensorflow as tf

from fedjax.training import logging


class LoggingTest(tf.test.TestCase):

  def test_log_no_root_dir(self):
    logger = logging.Logger()

    logger.log(
        writer_name='train', metric_name='loss', metric_value=4., round_num=0)

    self.assertEmpty(logger._summary_writers)

  def test_log_root_dir(self):
    root_dir = self.create_tempdir()
    logger = logging.Logger(root_dir)

    logger.log(
        writer_name='train', metric_name='loss', metric_value=4.1, round_num=0)
    logger.log(
        writer_name='eval', metric_name='loss', metric_value=5.3, round_num=0)

    self.assertCountEqual(['train', 'eval'], tf.io.gfile.listdir(root_dir))


if __name__ == '__main__':
  tf.test.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""FedJAX core."""

# Test utilities.
from fedjax.core import test_util

# Client trainer.
from fedjax.core.client_trainer import ClientTrainer
from fedjax.core.client_trainer import ControlVariateTrainer
from fedjax.core.client_trainer import ControlVariateTrainerState
from fedjax.core.client_trainer import DefaultClientTrainer
from fedjax.core.client_trainer import DefaultClientTrainerState
from fedjax.core.client_trainer import train_multiple_clients
from fedjax.core.client_trainer import train_single_client
# Dataset utilities.
from fedjax.core.dataset_util import ClientDataHParams
from fedjax.core.dataset_util import create_tf_dataset_for_clients
from fedjax.core.dataset_util import DatasetOrIterable
from fedjax.core.dataset_util import preprocess_tf_dataset
# Evaluation utilities
from fedjax.core.evaluation_util import aggregate_metrics
from fedjax.core.evaluation_util import evaluate_multiple_clients
from fedjax.core.evaluation_util import evaluate_single_client
# Federated algorithm.
from fedjax.core.federated_algorithm import FederatedAlgorithm
# Metrics.
from fedjax.core.metrics import accuracy_fn
from fedjax.core.metrics import CountMetric
from fedjax.core.metrics import cross_entropy_loss_fn
from fedjax.core.metrics import dataclass
from fedjax.core.metrics import masked_accuracy_fn
from fedjax.core.metrics import masked_accuracy_fn_with_logits_mask
from fedjax.core.metrics import masked_count
from fedjax.core.metrics import masked_cross_entropy_loss_fn
from fedjax.core.metrics import MeanMetric
from fedjax.core.metrics import Metric
from fedjax.core.metrics import oov_rate
from fedjax.core.metrics import sequence_length
from fedjax.core.metrics import truncation_rate
# JAX model.
from fedjax.core.model import create_model_from_haiku
from fedjax.core.model import create_model_from_stax
from fedjax.core.model import Model
# JAX optimizer.
from fedjax.core.optimizer import get_optimizer
from fedjax.core.optimizer import Optimizer
from fedjax.core.optimizer import OptimizerName
# Dataset prefetching.
from fedjax.core.prefetch import PrefetchClientDatasetsIterator
# Regularization
from fedjax.core.regularizers import L2Regularizer
from fedjax.core.regularizers import Regularizer
# Serialization.
from fedjax.core.serialization import load_state
from fedjax.core.serialization import save_state
# PyTree utilities.
from fedjax.core.tree_util import tree_map
from fedjax.core.tree_util import tree_mean
from fedjax.core.tree_util import tree_multimap
# Common type definitions.
from fedjax.core.typing import Batch
from fedjax.core.typing import FederatedData
from fedjax.core.typing import MetricResults
from fedjax.core.typing import OptState
from fedjax.core.typing import Params
from fedjax.core.typing import PRNGKey
from fedjax.core.typing import PRNGSequence
from fedjax.core.typing import Updates



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.training.federated_experiment."""

import collections
import os.path

from fedjax.core import test_util
from fedjax.training import federated_experiment
import tensorflow as tf


class FederatedExperimentTest(tf.test.TestCase):

  def test_run_federated_experiment(self):
    config = federated_experiment.FederatedExperimentConfig(
        root_dir=self.create_tempdir(),
        num_rounds=5,
        num_clients_per_round=2,
        checkpoint_frequency=3)
    num_examples = 3
    federated_algorithm = test_util.MockFederatedAlgorithm(
        num_examples=num_examples)

    state = federated_experiment.run_federated_experiment(
        config, federated_algorithm)

    self.assertEqual(
        state.count,
        config.num_rounds * config.num_clients_per_round * num_examples)
    self.assertTrue(
        os.path.exists(os.path.join(config.root_dir, 'checkpoint_00000003')))

  def test_run_federated_experiment_periodic_eval_fn_map(self):
    config = federated_experiment.FederatedExperimentConfig(
        root_dir=self.create_tempdir(),
        num_rounds=5,
        num_clients_per_round=2,
        eval_frequency=3)
    federated_algorithm = test_util.MockFederatedAlgorithm(num_examples=3)
    federated_data = federated_algorithm.federated_data
    model = federated_algorithm.model
    periodic_eval_fn_map = collections.OrderedDict(
        client_eval_1=federated_experiment.ClientEvaluationFn(
            federated_data, model, config),
        client_eval_2=federated_experiment.ClientEvaluationFn(
            federated_data, model, config),
        full_eval_1=federated_experiment.FullEvaluationFn(
            federated_data, model))

    federated_experiment.run_federated_experiment(
        config, federated_algorithm, periodic_eval_fn_map=periodic_eval_fn_map)

    self.assertTrue(
        os.path.exists(os.path.join(config.root_dir, 'client_eval_1')))
    self.assertTrue(
        os.path.exists(os.path.join(config.root_dir, 'client_eval_2')))
    self.assertTrue(
        os.path.exists(os.path.join(config.root_dir, 'full_eval_1')))

  def test_run_federated_experiment_final_eval_fn_map(self):
    config = federated_experiment.FederatedExperimentConfig(
        root_dir=self.create_tempdir(), num_rounds=5, num_clients_per_round=2)
    federated_algorithm = test_util.MockFederatedAlgorithm(num_examples=3)
    federated_data = federated_algorithm.federated_data
    model = federated_algorithm.model
    final_eval_fn_map = collections.OrderedDict(
        full_eval=federated_experiment.FullEvaluationFn(federated_data, model))

    federated_experiment.run_federated_experiment(
        config, federated_algorithm, final_eval_fn_map=final_eval_fn_map)

    self.assertTrue(
        os.path.exists(os.path.join(config.root_dir, 'full_eval.tsv')))

  def test_get_pseudo_random_state_with_random_seed(self):
    round_num = 100
    random_seed = 10

    random_state_1 = federated_experiment.get_pseudo_random_state(
        round_num, random_seed)
    client_ids_1 = random_state_1.choice(range(0, 100), size=5, replace=False)

    random_state_2 = federated_experiment.get_pseudo_random_state(
        round_num, random_seed)
    client_ids_2 = random_state_2.choice(range(0, 100), size=5, replace=False)

    self.assertAllEqual(client_ids_1, client_ids_2)

  def test_get_pseudo_random_state_without_random_seed(self):
    round_num = 100

    random_state_1 = federated_experiment.get_pseudo_random_state(round_num)
    client_ids_1 = random_state_1.choice(range(0, 100), size=5, replace=False)

    random_state_2 = federated_experiment.get_pseudo_random_state(round_num)
    client_ids_2 = random_state_2.choice(range(0, 100), size=5, replace=False)

    self.assertNotAllEqual(client_ids_1, client_ids_2)


if __name__ == '__main__':
  tf.test.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.aggregators.compression."""
from absl.testing import absltest

from fedjax import core
from fedjax.experimental.aggregators import compression
import jax.numpy as jnp
import jax.random as random
import numpy.testing as npt


class CompressionTest(absltest.TestCase):

  def test_num_leaves(self):
    _, model = core.test_util.create_toy_example(
        num_clients=10, num_clusters=4, num_classes=10, num_examples=5, seed=0)
    params = model.init_params(0)
    self.assertEqual(compression.num_leaves(params), 2)

  def test_num_params(self):
    _, model = core.test_util.create_toy_example(
        num_clients=10, num_clusters=4, num_classes=10, num_examples=5, seed=0)
    params = model.init_params(0)
    self.assertEqual(compression.num_params(params), 20)

  def test_binary_stochastic_quantize(self):
    # If the vector has only two distinct values, it should not change.
    v = jnp.array([0., 2., 2.])
    rng = random.PRNGKey(42)
    compressed_v = compression.binary_stochastic_quantize(v, rng)
    npt.assert_array_equal(compressed_v, v)

  def test_binary_stochastic_quantizer(self):
    num_classes = 10
    num_clients = 10
    num_examples = 5
    data, model = core.test_util.create_toy_example(
        num_clients=num_clients,
        num_clusters=4,
        num_classes=num_classes,
        num_examples=num_examples,
        seed=0)
    rng_seq = core.PRNGSequence(0)
    client_optimizer = core.get_optimizer(
        core.OptimizerName.SGD, learning_rate=0.1)
    client_trainer = core.DefaultClientTrainer(model, client_optimizer)
    init_params = model.init_params(0)
    client_outputs = core.train_multiple_clients(
        federated_data=data,
        client_ids=data.client_ids,
        client_trainer=client_trainer,
        init_client_trainer_state=client_trainer.init_state(init_params),
        rng_seq=rng_seq,
        client_data_hparams=core.ClientDataHParams(batch_size=10))

    def get_delta_params_and_weight(client_output):
      delta_params = core.tree_multimap(lambda a, b: a - b, init_params,
                                        client_output.params)
      return delta_params, client_output.num_examples

    delta_params_and_weight = map(get_delta_params_and_weight, client_outputs)

    default_aggregator = compression.BinaryStochasticQuantizer()
    init_aggregator_state = default_aggregator.init_state()
    rng_seq2 = core.PRNGSequence(1)
    _, new_state = default_aggregator.aggregate(init_aggregator_state,
                                                delta_params_and_weight,
                                                rng_seq2)
    self.assertEqual(new_state.total_weight, num_clients * num_examples)
    self.assertEqual(new_state.num_bits, 1480.0)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.experimental.client_datasets."""

import itertools

from absl.testing import absltest
from fedjax.experimental import client_datasets
import numpy as np
import numpy.testing as npt

# pylint: disable=g-long-lambda


class ExamplesTest(absltest.TestCase):

  def test_num_examples(self):
    with self.subTest('consistent rows'):
      self.assertEqual(
          client_datasets.num_examples({'pixels': np.zeros([10, 20, 30])}), 10)
    with self.subTest('inconsistent rows'):
      with self.assertRaises(ValueError):
        client_datasets.num_examples({
            'a': np.zeros([10, 20]),
            'b': np.zeros([20, 10])
        })

  def test_slice_examples(self):
    examples = {'a': np.arange(10), 'b': np.arange(20).reshape([10, 2])}

    with self.subTest('Slicing like [3:]'):
      sliced = client_datasets.slice_examples(examples, slice(3))
      self.assertCountEqual(sliced, ['a', 'b'])
      npt.assert_array_equal(sliced['a'], [0, 1, 2])
      npt.assert_array_equal(sliced['b'], [[0, 1], [2, 3], [4, 5]])

    with self.subTest('Slicing like [3:6]'):
      sliced = client_datasets.slice_examples(examples, slice(3, 6))
      self.assertCountEqual(sliced, ['a', 'b'])
      npt.assert_array_equal(sliced['a'], [3, 4, 5])
      npt.assert_array_equal(sliced['b'], [[6, 7], [8, 9], [10, 11]])

    with self.subTest('Slicing like [3:6:2]'):
      sliced = client_datasets.slice_examples(examples, slice(3, 6, 2))
      self.assertCountEqual(sliced, ['a', 'b'])
      npt.assert_array_equal(sliced['a'], [3, 5])
      npt.assert_array_equal(sliced['b'], [[6, 7], [10, 11]])

  def test_pad_examples(self):
    examples = {'a': np.arange(5), 'b': np.arange(10).reshape([5, 2])}
    with self.subTest('mask_key already used'):
      with self.assertRaises(ValueError):
        client_datasets.pad_examples(examples, size=8, mask_key='a')
    with self.subTest('size too small'):
      with self.assertRaises(ValueError):
        client_datasets.pad_examples(examples, size=4, mask_key='mask')
    with self.subTest('no padding needed'):
      self.assertIs(client_datasets.pad_examples(examples, 5, 'mask'), examples)
    with self.subTest('padding needed'):
      padded = client_datasets.pad_examples(examples, 6, 'mask')
      self.assertCountEqual(padded, ['a', 'b', 'mask'])
      npt.assert_array_equal(padded['a'], [0, 1, 2, 3, 4, 0])
      npt.assert_array_equal(padded['b'],
                             [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9], [0, 0]])
      npt.assert_array_equal(padded['mask'],
                             [True, True, True, True, True, False])


class PreprocessorTest(absltest.TestCase):

  def test_preprocessor(self):
    preprocessor = client_datasets.Preprocessor([
        # Flattens `pixels`.
        lambda x: {
            **x, 'pixels': x['pixels'].reshape([-1, 28 * 28])
        },
        # Introduce `binary_label`.
        lambda x: {
            **x, 'binary_label': x['label'] % 2
        },
    ])
    fake_emnist = {
        'pixels': np.random.uniform(size=(10, 28, 28)),
        'label': np.random.randint(10, size=(10,))
    }

    with self.subTest('2 step preprocessing'):
      result = preprocessor(fake_emnist)
      self.assertCountEqual(result, ['pixels', 'label', 'binary_label'])
      npt.assert_array_equal(result['pixels'],
                             fake_emnist['pixels'].reshape([-1, 28 * 28]))
      self.assertIs(result['label'], fake_emnist['label'])
      npt.assert_array_equal(result['binary_label'], fake_emnist['label'] % 2)

  def test_append(self):
    preprocessor = client_datasets.Preprocessor([
        # Flattens `pixels`.
        lambda x: {
            **x, 'pixels': x['pixels'].reshape([-1, 28 * 28])
        },
        # Introduce `binary_label`.
        lambda x: {
            **x, 'binary_label': x['label'] % 2
        },
    ])
    new_preprocessor = preprocessor.append(lambda x: {
        **x, 'sum_pixels': np.sum(x['pixels'], axis=1)
    })
    self.assertIsNot(new_preprocessor, preprocessor)
    fake_emnist = {
        'pixels': np.random.uniform(size=(10, 28, 28)),
        'label': np.random.randint(10, size=(10,))
    }
    result = new_preprocessor(fake_emnist)
    self.assertCountEqual(result,
                          ['pixels', 'label', 'binary_label', 'sum_pixels'])
    npt.assert_array_equal(result['pixels'],
                           fake_emnist['pixels'].reshape([-1, 28 * 28]))
    self.assertIs(result['label'], fake_emnist['label'])
    npt.assert_array_equal(result['binary_label'], fake_emnist['label'] % 2)
    npt.assert_allclose(result['sum_pixels'],
                        np.sum(fake_emnist['pixels'], axis=(1, 2)))


class ClientDatasetTest(absltest.TestCase):

  def test_inconsistent_shape(self):
    with self.assertRaises(ValueError):
      client_datasets.ClientDataset({'a': np.arange(5), 'b': np.arange(4)})

  def test_len(self):
    examples = {'a': np.arange(10)}
    self.assertLen(client_datasets.ClientDataset(examples), 10)

  def test_batch(self):
    d = client_datasets.ClientDataset(
        {
            'a': np.arange(5),
            'b': np.arange(10).reshape([5, 2])
        }, client_datasets.Preprocessor([lambda x: {
            **x, 'a': 2 * x['a']
        }]))
    view = d.batch(3)
    # `view` should be repeatedly iterable.
    for _ in range(2):
      batches = list(view)
      self.assertLen(batches, 2)
      batch_0, batch_1 = batches
      self.assertCountEqual(batch_0, ['a', 'b'])
      npt.assert_array_equal(batch_0['a'], [0, 2, 4])
      npt.assert_array_equal(batch_0['b'], [[0, 1], [2, 3], [4, 5]])
      self.assertCountEqual(batch_1, ['a', 'b'])
      npt.assert_array_equal(batch_1['a'], [6, 8])
      npt.assert_array_equal(batch_1['b'], [[6, 7], [8, 9]])

  def test_pick_final_batch_size(self):
    self.assertEqual(client_datasets._pick_final_batch_size(16, 8, 4), 8)
    self.assertEqual(client_datasets._pick_final_batch_size(10, 8, 1), 8)
    self.assertEqual(client_datasets._pick_final_batch_size(10, 8, 2), 4)
    self.assertEqual(client_datasets._pick_final_batch_size(10, 8, 3), 2)
    self.assertEqual(client_datasets._pick_final_batch_size(10, 8, 4), 2)
    self.assertEqual(client_datasets._pick_final_batch_size(10, 8, 5), 2)
    self.assertEqual(client_datasets._pick_final_batch_size(9, 8, 1), 8)
    self.assertEqual(client_datasets._pick_final_batch_size(9, 8, 2), 4)
    self.assertEqual(client_datasets._pick_final_batch_size(9, 8, 3), 2)
    self.assertEqual(client_datasets._pick_final_batch_size(9, 8, 4), 1)

  def test_batch_padding(self):
    d = client_datasets.ClientDataset(
        {
            'a': np.arange(5),
            'b': np.arange(10).reshape([5, 2])
        }, client_datasets.Preprocessor([lambda x: {
            **x, 'a': 2 * x['a']
        }]))
    view = d.batch(3, num_batch_size_buckets=1, mask_key='M')
    # `view` should be repeatedly iterable.
    for _ in range(2):
      batches = list(view)
      self.assertLen(batches, 2)
      batch_0, batch_1 = batches
      self.assertCountEqual(batch_0, ['a', 'b'])
      npt.assert_array_equal(batch_0['a'], [0, 2, 4])
      npt.assert_array_equal(batch_0['b'], [[0, 1], [2, 3], [4, 5]])
      self.assertCountEqual(batch_1, ['a', 'b', 'M'])
      npt.assert_array_equal(batch_1['a'], [6, 8, 0])
      npt.assert_array_equal(batch_1['b'], [[6, 7], [8, 9], [0, 0]])
      npt.assert_array_equal(batch_1['M'], [True, True, False])

  def test_shuffle_repeat_batch(self):
    d = client_datasets.ClientDataset(
        {
            'a': np.arange(5),
            'b': np.arange(10).reshape([5, 2])
        }, client_datasets.Preprocessor([lambda x: {
            **x, 'a': 2 * x['a']
        }]))
    # Number of batches under different num_epochs/num_steps combinations.
    with self.subTest('repeating'):
      self.assertLen(list(d.shuffle_repeat_batch(5, num_epochs=1)), 1)
      self.assertLen(list(d.shuffle_repeat_batch(3, num_epochs=1)), 2)
      self.assertLen(list(d.shuffle_repeat_batch(1, num_epochs=1)), 5)
      self.assertLen(list(d.shuffle_repeat_batch(5, num_steps=4)), 4)
      self.assertLen(list(d.shuffle_repeat_batch(3, num_steps=4)), 4)
      self.assertLen(list(d.shuffle_repeat_batch(1, num_steps=4)), 4)
      self.assertLen(
          list(d.shuffle_repeat_batch(5, num_epochs=1, num_steps=4)), 1)
      self.assertLen(
          list(d.shuffle_repeat_batch(3, num_epochs=1, num_steps=4)), 2)
      self.assertLen(
          list(d.shuffle_repeat_batch(1, num_epochs=1, num_steps=4)), 4)
      # 100 is as good as forever.
      self.assertLen(
          list(itertools.islice(d.shuffle_repeat_batch(3), 100)), 100)

    # Check proper shuffling.
    with self.subTest('shuffling'):
      view = d.shuffle_repeat_batch(3, num_steps=4, seed=1)
      # `view` should be repeatedly iterable.
      for _ in range(2):
        batches = list(view)
        self.assertLen(batches, 4)
        self.assertCountEqual(batches[0], ['a', 'b'])
        npt.assert_array_equal(batches[0]['a'], [4, 2, 8])
        npt.assert_array_equal(batches[0]['b'], [[4, 5], [2, 3], [8, 9]])
        self.assertCountEqual(batches[1], ['a', 'b'])
        npt.assert_array_equal(batches[1]['a'], [0, 6, 4])
        npt.assert_array_equal(batches[1]['b'], [[0, 1], [6, 7], [4, 5]])
        self.assertCountEqual(batches[2], ['a', 'b'])
        npt.assert_array_equal(batches[2]['a'], [8, 6, 0])
        npt.assert_array_equal(batches[2]['b'], [[8, 9], [6, 7], [0, 1]])
        self.assertCountEqual(batches[3], ['a', 'b'])
        npt.assert_array_equal(batches[3]['a'], [2, 6, 0])
        npt.assert_array_equal(batches[3]['b'], [[2, 3], [6, 7], [0, 1]])

  def test_slice(self):
    d = client_datasets.ClientDataset(
        {
            'a': np.arange(5),
            'b': np.arange(10).reshape([5, 2])
        }, client_datasets.Preprocessor([lambda x: {
            **x, 'a': 2 * x['a']
        }]))

    with self.subTest('slice [:3]'):
      sliced = d[:3]
      batch = next(iter(sliced.batch(3)))
      self.assertCountEqual(batch, ['a', 'b'])
      npt.assert_array_equal(batch['a'], [0, 2, 4])
      npt.assert_array_equal(batch['b'], [[0, 1], [2, 3], [4, 5]])

    with self.subTest('slice [-3:]'):
      sliced = d[-3:]
      batch = next(iter(sliced.batch(3)))
      self.assertCountEqual(batch, ['a', 'b'])
      npt.assert_array_equal(batch['a'], [4, 6, 8])
      npt.assert_array_equal(batch['b'], [[4, 5], [6, 7], [8, 9]])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""FederatedData interface for providing access to a federated dataset."""

import abc
import itertools
import random
from typing import Any, Callable, Iterable, Iterator, Optional, Tuple

from fedjax.experimental import client_datasets

# A client id is simply some binary bytes.
ClientId = bytes


class FederatedData(abc.ABC):
  """FederatedData interface for providing access to a federated dataset.

  A `FederatedData` object serves as a mapping from client ids to client
  datasets and client metadata.

  ##  Access methods with better I/O efficiency

  For large federated datasets, it is not feasible to load all client datasets
  into memory at once (whereas loading a single client dataset is assumed to be
  feasible). Different implementations exist for different on disk storage
  formats. Since sequential read is much faster than random read for most
  storage technologies, `FederatedData` provides two types of methods for
  accessing client datasets,

  1.  `clients()` and `shuffled_clients()` are sequential read friendly, and
  thus recommended whenever appropriate.
  2.  `get_clients()` requires random read, but prefetching is possible. This
  should be preferred over `get_client()`.
  3.  `get_client()` is usually the slowest way of accessing client datasets,
  and is mostly intended for interactive exploration of a small number of
  clients.

  ##  Preprocessing

  `ClientDataset`s produced by `FederatedData` can hold a `Preprocessor`,
  customizable via `preprocess_batch()`. Additionally, another "client" level
  `Preprocessor`, customizable via `preprocess_client()`, can be used to apply
  transformations on examples from the entire client dataset before a
  ClientDataset is constructed.
  """

  @abc.abstractmethod
  def slice(self,
            start: Optional[ClientId] = None,
            stop: Optional[ClientId] = None) -> 'FederatedData':
    """Returns a new FederatedData restricted to client ids in the given range.

    The returned FederatedData includes clients whose ids are,
    -   Greater than or equal to `start` when `start` is not None;
    -   Less than `stop` when `stop` is not None.

    Args:
      start: Start of client id range.
      stop: Stop of client id range.

    Returns:
      FederatedData.
    """

  # Client metadata access.

  @abc.abstractmethod
  def num_clients(self) -> int:
    """Returns the number of clients.

    If it is too expensive or otherwise impossible to obtain the result, an
    implementation may raise an exception.
    """

  @abc.abstractmethod
  def client_ids(self) -> Iterator[ClientId]:
    """Returns an iterable of client ids as bytes.

    There is no requirement on the order of iteration.
    """

  @abc.abstractmethod
  def client_sizes(self) -> Iterator[Tuple[ClientId, int]]:
    """Returns an iterable of all (client id, client size) pairs.

    This is often more efficient than making multiple client_size() calls. There
    is no requirement on the order of iteration.
    """

  @abc.abstractmethod
  def client_size(self, client_id: ClientId) -> int:
    """Returns the number of examples in a client dataset."""

  # Client access.

  @abc.abstractmethod
  def clients(self) -> Iterator[Tuple[ClientId, client_datasets.ClientDataset]]:
    """Iterates over clients in a deterministic order.

    Implementation can choose whatever order that makes iteration efficient.
    """

  @abc.abstractmethod
  def shuffled_clients(
      self,
      buffer_size: int,
      seed: Optional[int] = None
  ) -> Iterator[Tuple[ClientId, client_datasets.ClientDataset]]:
    """Iterates over clients with a repeated buffered shuffling.

    Shuffling should use a buffer with a size of at least `buffer_size` clients.
    The iteration should repeat forever, with usually a different order in each
    pass.

    Args:
      buffer_size: Buffer size for shuffling.
      seed: Optional random number generator seed.

    Returns:
      Iterator.
    """

  @abc.abstractmethod
  def get_clients(
      self, client_ids: Iterable[ClientId]
  ) -> Iterator[Tuple[ClientId, client_datasets.ClientDataset]]:
    """Gets multiple clients in order with one call.

    Clients are returned in the order of `client_ids`.

    Args:
      client_ids: Client ids to load.

    Returns:
      Iterator.
    """

  @abc.abstractmethod
  def get_client(self, client_id: ClientId) -> client_datasets.ClientDataset:
    """Gets one single client dataset.

    Prefer clients(), shuffled_clients(), or get_clients() when possible.

    Args:
      client_id: Client id to load.

    Returns:
      The corresponding ClientDataset.
    """

  # Preprocessing.

  @abc.abstractmethod
  def preprocess_client(
      self, fn: Callable[[client_datasets.Examples], client_datasets.Examples]
  ) -> 'FederatedData':
    """Registers a preprocessing function to be called on all examples of a client before passing them to construct a ClientDataset."""

  @abc.abstractmethod
  def preprocess_batch(
      self, fn: Callable[[client_datasets.Examples], client_datasets.Examples]
  ) -> 'FederatedData':
    """Registers a preprocessing function to be called after batching in ClientDatasets."""


# Utility functions useful when implementing FederatedData.


def buffered_shuffle(source: Iterable[Any], buffer_size: int,
                     rng: random.Random) -> Iterator[Any]:
  """Shuffles an iterable via buffered shuffling."""
  it = iter(source)
  buf = list(itertools.islice(it, buffer_size))
  rng.shuffle(buf)
  for i in it:
    r, buf[0] = buf[0], i
    swap = rng.randrange(buffer_size)
    if swap < buffer_size - 1:
      buf[swap], buf[0] = buf[0], buf[swap]
    yield r
  for i in buf:
    yield i



"""Tests for fedjax.core.prefetch."""

from concurrent import futures
import time

from absl import flags
from absl.testing import absltest
from fedjax.core import prefetch
import tensorflow as tf

FLAGS = flags.FLAGS
flags.DEFINE_bool(
    'run_timing_test', False, 'Whether to run tests that check timing results. '
    'Such tests are important for verifying performance characteristics but '
    'often fail under some testing tools such as thread sanitizer.')


class AsyncTFDatasetIteratorTest(absltest.TestCase):

  def test_one_pass(self):
    values = []
    with futures.ThreadPoolExecutor(1) as executor:
      for i in prefetch.AsyncTFDatasetIterator(
          executor, lambda: tf.data.Dataset.range(5)):
        values.append(i)
    self.assertEqual(values, [0, 1, 2, 3, 4])

  def test_dataset_fn_exception(self):

    def dataset_fn():
      raise ValueError

    with futures.ThreadPoolExecutor(1) as executor:
      it = prefetch.AsyncTFDatasetIterator(executor, dataset_fn)
      with self.assertRaises(ValueError):
        next(it)

  def test_elapsed_time(self):
    if not FLAGS.run_timing_test:
      return

    # This test may fail if we do not build with -c opt.
    create_time = 0.5
    next_time = 0.2

    def work(x):
      time.sleep(next_time)
      return x

    def tf_work(x):
      return tf.py_function(work, [x], tf.int64)

    def dataset_fn():
      time.sleep(create_time)
      return tf.data.Dataset.range(5).map(tf_work)

    with futures.ThreadPoolExecutor(1) as executor:
      start = time.perf_counter()
      it = prefetch.AsyncTFDatasetIterator(executor, dataset_fn)
      time.sleep(create_time + next_time)
      next(it)
      time_to_first_value = time.perf_counter() - start
    # (create_time + next_time) * 2 is how long it takes to execute the code
    # above synchronously.
    self.assertLess(time_to_first_value, (create_time + next_time) * 2)


class PrefetchClientDatasetsIteratorTest(absltest.TestCase):

  def test_one_pass(self):

    class FakeFederatedData:

      def __init__(self, offset):
        self._offset = offset

      def create_tf_dataset_for_client(self, client_id):
        return tf.data.Dataset.range(int(client_id) + self._offset)

    federated_data = (FakeFederatedData(0), FakeFederatedData(1))
    client_ids = ['0', '1', '2']
    values = list(
        prefetch.PrefetchClientDatasetsIterator(federated_data, client_ids))
    self.assertLen(values, 3)
    id0, (it00, it01) = values[0]
    self.assertEqual(id0, '0')
    self.assertEqual(list(it00), [])
    self.assertEqual(list(it01), [0])
    id1, (it10, it11) = values[1]
    self.assertEqual(id1, '1')
    self.assertEqual(list(it10), [0])
    self.assertEqual(list(it11), [0, 1])
    id2, (it20, it21) = values[2]
    self.assertEqual(id2, '2')
    self.assertEqual(list(it20), [0, 1])
    self.assertEqual(list(it21), [0, 1, 2])


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Simple federated averaging implementation.

Identical to the implementation at //fedjax/algorithms/fed_avg.py

Based on the paper:

Communication-Efficient Learning of Deep Networks from Decentralized Data
    H. Brendan McMahan, Eider Moore, Daniel Ramage,
    Seth Hampson, Blaise Aguera y Arcas. AISTATS 2017.
    https://arxiv.org/abs/1602.05629
"""

from typing import List, NamedTuple

import fedjax


class SimpleFedAvgState(NamedTuple):
  """State of server passed between rounds.

  Attributes:
    params: A pytree representing the server model parameters.
    server_opt_state: A pytree representing the server optimizer state.
  """
  params: fedjax.Params
  server_opt_state: fedjax.OptState


class SimpleFedAvgHParams(NamedTuple):
  """Hyperparameters for federated averaging.

  Attributes:
    train_data_hparams: Hyperparameters for training data preparation.
  """
  train_data_hparams: fedjax.ClientDataHParams


class SimpleFedAvg(fedjax.FederatedAlgorithm):
  """Simple federated averaging algorithm."""

  def __init__(self, federated_data: fedjax.FederatedData, model: fedjax.Model,
               client_optimizer: fedjax.Optimizer,
               server_optimizer: fedjax.Optimizer, hparams: SimpleFedAvgHParams,
               rng_seq: fedjax.PRNGSequence):
    """Initializes SimpleFedAvg algorithm.

    Args:
      federated_data: Federated data separated per client.
      model: Model implementation.
      client_optimizer: Client optimizer.
      server_optimizer: Server optimizer.
      hparams: Hyperparameters for federated averaging.
      rng_seq: Iterator of JAX random keys.
    """
    self._federated_data = federated_data
    self._model = model
    self._client_optimizer = client_optimizer
    self._server_optimizer = server_optimizer
    self._hparams = hparams
    self._rng_seq = rng_seq
    self._client_trainer = fedjax.DefaultClientTrainer(model, client_optimizer)

  @property
  def federated_data(self) -> fedjax.FederatedData:
    return self._federated_data

  @property
  def model(self) -> fedjax.Model:
    return self._model

  def init_state(self) -> SimpleFedAvgState:
    params = self._model.init_params(next(self._rng_seq))
    server_opt_state = self._server_optimizer.init_fn(params)
    return SimpleFedAvgState(params=params, server_opt_state=server_opt_state)

  def run_round(self, state: SimpleFedAvgState,
                client_ids: List[str]) -> SimpleFedAvgState:
    """Runs one round of federated averaging."""
    # Train model per client.
    client_outputs = fedjax.train_multiple_clients(
        federated_data=self.federated_data,
        client_ids=client_ids,
        client_trainer=self._client_trainer,
        init_client_trainer_state=self._client_trainer.init_state(state.params),
        rng_seq=self._rng_seq,
        client_data_hparams=self._hparams.train_data_hparams)

    # Weighted average of param delta across clients.
    def get_delta_params_and_weight(client_output):
      delta_params = fedjax.tree_multimap(lambda a, b: a - b, state.params,
                                          client_output.params)
      return delta_params, client_output.num_examples

    delta_params_and_weight = map(get_delta_params_and_weight, client_outputs)
    delta_params = fedjax.tree_mean(delta_params_and_weight)

    # Server state update.
    updates, server_opt_state = self._server_optimizer.update_fn(
        delta_params, state.server_opt_state)
    params = self._server_optimizer.apply_updates(state.params, updates)
    return SimpleFedAvgState(params=params, server_opt_state=server_opt_state)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Algorithms for model update compression."""

from typing import Iterable, NamedTuple, Tuple, TypeVar

from fedjax import core
from fedjax.experimental.aggregators import aggregator
import jax
import jax.numpy as jnp

W = TypeVar('W')


class CompressionState(NamedTuple):
  """State of default aggregator passed between rounds.

  Attributes:
    total_weight: the number of samples.
    num_bits: number of bits transmitted.
  """
  total_weight: float
  num_bits: float


def binary_stochastic_quantize(v, rng):
  v_min = jnp.amin(v)
  v_max = jnp.amax(v)
  v = (v - v_min) / (v_max - v_min + 1e-15)
  rand = jax.random.uniform(key=rng, shape=v.shape)
  return jnp.where(rand > v, v_min, v_max)


def num_leaves(pytree):
  return len(jax.tree_util.tree_leaves(pytree))


def num_params(pytree):
  x = jax.tree_util.tree_leaves(pytree)
  return sum(i.size for i in x)


class BinaryStochasticQuantizer(aggregator.Aggregator):
  """Binary stochastic algorithm in https://arxiv.org/pdf/1611.00429.pdf."""

  def init_state(self):
    return CompressionState(total_weight=0.0, num_bits=0.0)

  def aggregate(self, aggregator_state: CompressionState,
                params_and_weight: Iterable[Tuple[W, float]],
                rng_seq: core.PRNGSequence) -> Tuple[CompressionState, W]:
    """Returns (weighted) mean of input trees and weights.

    Args:
      aggregator_state: state of the input.
      params_and_weight: Iterable of tuples of pytrees and associated weights.
      rng_seq: Random sequence.

    Returns:
      New state, (Weighted) mean of input trees and weights.
    """

    def quantize_params_and_weight(param_weight_rng):
      param_weight, rng = param_weight_rng
      param, weight = param_weight

      def binary_stochastic_quantize_with_rng(param):
        return binary_stochastic_quantize(param, rng)
      return jax.tree_map(binary_stochastic_quantize_with_rng, param), weight

    # TODO(theertha): avoid the need to copying the entire sequence to memory.

    params_and_weight_rng = zip(params_and_weight, rng_seq)
    quantized_p_and_w = map(quantize_params_and_weight,
                            params_and_weight_rng)
    quantized_p_and_w = list(quantized_p_and_w)
    weights = [weight for params, weight in quantized_p_and_w]
    new_weight = sum(weights)
    params = [params for params, weight in quantized_p_and_w]
    total_num_floats = sum([num_leaves(param) for param in params])
    total_num_params = sum([num_params(param) for param in params])
    # 32 bits for every float used and one bit for every parameter.
    new_bits = total_num_params + 64 * total_num_floats
    new_state = CompressionState(
        total_weight=aggregator_state.total_weight + new_weight,
        num_bits=aggregator_state.num_bits + new_bits)
    return core.tree_mean(quantized_p_and_w), new_state



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Federated learning algorithm implementations."""

from fedjax.algorithms.agnostic_fed_avg import AgnosticFedAvg
from fedjax.algorithms.agnostic_fed_avg import AgnosticFedAvgHParams
from fedjax.algorithms.agnostic_fed_avg import AgnosticFedAvgState
from fedjax.algorithms.agnostic_fed_avg import DomainAlgorithm
from fedjax.algorithms.fed_avg import FedAvg
from fedjax.algorithms.fed_avg import FedAvgHParams
from fedjax.algorithms.fed_avg import FedAvgState
from fedjax.algorithms.hyp_cluster import evaluate_multiple_clients_with_clusters
from fedjax.algorithms.hyp_cluster import HypCluster
from fedjax.algorithms.hyp_cluster import HypClusterHParams
from fedjax.algorithms.hyp_cluster import HypClusterState
from fedjax.algorithms.mime import Mime
from fedjax.algorithms.mime_lite import MimeLite
from fedjax.algorithms.mime_lite import MimeLiteClientTrainer
from fedjax.algorithms.mime_lite import MimeLiteHParams
from fedjax.algorithms.mime_lite import MimeLiteState



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Install fedjax."""

from setuptools import find_namespace_packages
from setuptools import setup

__version__ = None

with open('fedjax/version.py') as f:
  exec(f.read(), globals())

setup(
    name='fedjax',
    version=__version__,
    url='https://github.com/google/fedjax',
    license='Apache 2.0',
    author='FedJAX Team',
    author_email='no-reply@google.com',
    description=('Federated learning simulation with JAX.'),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    keywords='federated python machine learning',
    packages=find_namespace_packages(exclude=['*_benchmark.py', '*_test.py']),
    install_requires=[
        'dm-haiku',
        'immutabledict',
        'jax',
        'jaxlib',
        'msgpack',
        'optax',
        'tensorflow-federated',
    ],
    # TensorFlow doesn't yet support Python 3.9.
    python_requires='>=3.6, <3.9',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Methods for checkpointing."""

import os.path
import re
from typing import Any, List, Optional, Tuple

from fedjax import core
import tensorflow as tf

_CHECKPOINT_PREFIX = 'checkpoint_'


def _get_checkpoint_paths(base_path: str) -> List[str]:
  """Returns all checkpoint paths present."""
  pattern = base_path + r'[0-9]{8}$'
  checkpoint_paths = []
  for path in tf.io.gfile.glob(base_path + '*'):
    if re.match(pattern, path):
      checkpoint_paths.append(path)

  def sort_key(path):
    return int(path.split(base_path)[-1])

  return sorted(checkpoint_paths, key=sort_key)


def load_latest_checkpoint(root_dir: str) -> Optional[Tuple[Any, int]]:
  """Loads latest checkpoint and round number."""
  base_path = os.path.join(root_dir, _CHECKPOINT_PREFIX)
  all_checkpoint_paths = _get_checkpoint_paths(base_path)
  if all_checkpoint_paths:
    latest_checkpoint_path = all_checkpoint_paths[-1]
    latest_round_num = int(latest_checkpoint_path.split(base_path)[-1])
    latest_state = core.load_state(latest_checkpoint_path)
    return latest_state, latest_round_num


def save_checkpoint(root_dir: str,
                    state: Any,
                    round_num: int = 0,
                    keep: int = 1):
  """Saves checkpoint and cleans up old checkpoints."""
  base_path = os.path.join(root_dir, _CHECKPOINT_PREFIX)
  checkpoint_path = f'{base_path}{round_num:08d}'
  core.save_state(state, checkpoint_path)
  remove_checkpoint_paths = _get_checkpoint_paths(base_path)[:-keep]
  for path in remove_checkpoint_paths:
    tf.io.gfile.remove(path)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.metrics."""

from absl.testing import parameterized
from fedjax.core import metrics
import jax.numpy as jnp
import tensorflow as tf


class MetricsTest(tf.test.TestCase, parameterized.TestCase):

  @parameterized.named_parameters(
      {
          'testcase_name': 'rank_1',
          'targets': [1, 0, 1],
          # (batch_size, num_classes).
          'preds': [[1.2, 0.4], [2.3, 0.1], [0.3, 3.2]],
          'expected_loss': 0.44324896,
      },
      {
          'testcase_name': 'rank_2',
          'targets': [[1, 0], [0, 1], [0, 1]],
          # (batch_size, time_steps, num_classes).
          'preds': [[[2.3, 0.2], [1.2, 2.3]], [[0.4, 0.6], [0.1, 1.2]],
                    [[3.2, 2.1], [0.1, 2.0]]],
          'expected_loss': 0.8525085,
      })
  def test_cross_entropy_loss_fn(self, targets, preds, expected_loss):
    targets, preds = jnp.array(targets), jnp.array(preds)
    loss = metrics.cross_entropy_loss_fn(targets, preds)
    self.assertAlmostEqual(loss.result(), expected_loss)

  @parameterized.named_parameters(
      {
          'testcase_name': 'rank_1',
          'targets': [1, 0, 1],
          # (batch_size, num_classes).
          'preds': [[1.2, 0.4], [2.3, 0.1], [0.3, 3.2]],
          'mask_values': (0,),
          'expected_loss': 0.612331725,
      },
      {
          'testcase_name': 'rank_2',
          'targets': [[1, 0], [0, 1], [0, 1]],
          # (batch_size, time_steps, num_classes).
          'preds': [[[2.3, 0.2], [1.2, 2.3]], [[0.4, 0.6], [0.1, 1.2]],
                    [[3.2, 2.1], [0.1, 2.0]]],
          'mask_values': (0,),
          'expected_loss': 0.880747,
      },
      {
          'testcase_name': 'rank_2_multi_mask',
          'targets': [[1, 0], [0, 2], [0, 1]],
          # (batch_size, time_steps, num_classes).
          'preds': [[[2.3, 0.2], [1.2, 2.3]], [[0.4, 0.6], [0.1, 1.2]],
                    [[3.2, 2.1], [0.1, 2.0]]],
          'mask_values': (0, 2),
          'expected_loss': 1.17745305,
      })
  def test_masked_cross_entropy_loss_fn(self, targets, preds, mask_values,
                                        expected_loss):
    targets, preds = jnp.array(targets), jnp.array(preds)
    loss = metrics.masked_cross_entropy_loss_fn(targets, preds, mask_values)
    self.assertAllClose(loss.result(), expected_loss)

  @parameterized.named_parameters(
      {
          'testcase_name': 'rank_1',
          'targets': [1, 0, 2, 0],
          'preds': [[0, 1, 0], [1, 0, 0], [0, 0, 1], [0, 1, 0]],
          'expected_accuracy': 3. / 4.,
      }, {
          'testcase_name': 'rank_2',
          'targets': [[1, 0], [1, 1], [1, 1], [0, 2]],
          'preds': [[[0, 1, 0], [1, 0, 0]], [[0, 0, 1], [0, 0, 1]],
                    [[1, 0, 0], [0, 1, 0]], [[1, 0, 0], [0, 1, 0]]],
          'expected_accuracy': 4. / 8.,
      })
  def test_accuracy_fn(self, targets, preds, expected_accuracy):
    targets, preds = jnp.array(targets), jnp.array(preds)
    accuracy = metrics.accuracy_fn(targets, preds)
    self.assertAlmostEqual(accuracy.result(), expected_accuracy)

  @parameterized.named_parameters(
      {
          'testcase_name': 'rank_1',
          'targets': [1, 0, 2, 0],
          'preds': [[0, 1, 0], [1, 0, 0], [0, 0, 1], [0, 1, 0]],
          'mask_values': (0,),
          'expected_accuracy': 2. / 2.,
      }, {
          'testcase_name': 'rank_2',
          'targets': [[1, 0], [1, 1], [1, 1], [0, 2]],
          'preds': [[[0, 1, 0], [1, 0, 0]], [[0, 0, 1], [0, 0, 1]],
                    [[1, 0, 0], [0, 1, 0]], [[1, 0, 0], [0, 1, 0]]],
          'mask_values': (0,),
          'expected_accuracy': 2. / 6.,
      }, {
          'testcase_name': 'rank_2_multi_mask',
          'targets': [[1, 0], [1, 1], [2, 1], [0, 2]],
          'preds': [[[0, 1, 0], [1, 0, 0]], [[0, 0, 1], [0, 0, 1]],
                    [[1, 0, 0], [0, 1, 0]], [[1, 0, 0], [0, 1, 0]]],
          'mask_values': (0, 2),
          'expected_accuracy': 2. / 4.,
      })
  def test_masked_accuracy_fn(self, targets, preds, mask_values,
                              expected_accuracy):
    targets, preds = jnp.array(targets), jnp.array(preds)
    accuracy = metrics.masked_accuracy_fn(targets, preds, mask_values)
    self.assertAlmostEqual(accuracy.result(), expected_accuracy)

  @parameterized.named_parameters(
      {
          'testcase_name': 'rank_1',
          'targets': [1, 0, 2, 0],
          'preds': [[0, 1, 0], [1, 0, 0], [0, 0, 1], [0, 1, 0]],
          'mask_values': (0,),
          'logits_mask': [0, jnp.NINF, 0],
          'expected_accuracy': 1. / 2.,
      }, {
          'testcase_name': 'rank_2',
          'targets': [[1, 0], [1, 1], [1, 1], [0, 2]],
          'preds': [[[0, 1, 0], [1, 0, 0]], [[0, 0, 1], [0, 0, 1]],
                    [[1, 0, 0], [0, 1, 0]], [[1, 0, 0], [0, 1, 0]]],
          'mask_values': (0,),
          'logits_mask': [0, jnp.NINF, 0],
          'expected_accuracy': 0. / 6.,
      }, {
          'testcase_name': 'rank_2_multi_mask',
          'targets': [[1, 0], [1, 1], [2, 1], [0, 2]],
          'preds': [[[0, 1, 0], [1, 0, 0]], [[0, 0, 1], [0, 0, 1]],
                    [[1, 0, 0], [0, 1, 0]], [[1, 0, 0], [0, 1, 0]]],
          'mask_values': (0, 2),
          'logits_mask': [0, jnp.NINF, 0],
          'expected_accuracy': 0. / 4.,
      })
  def test_masked_accuracy_fn_with_logits_mask(self, targets, preds,
                                               mask_values, logits_mask,
                                               expected_accuracy):
    targets, preds, logits_mask = jnp.array(targets), jnp.array(
        preds), jnp.array(logits_mask)
    accuracy = metrics.masked_accuracy_fn_with_logits_mask(
        targets, preds, logits_mask, mask_values)
    self.assertAlmostEqual(accuracy.result(), expected_accuracy)

  def test_str_version_of_metric(self):
    metric = metrics.MeanMetric(total=2, count=4)
    self.assertEqual('MeanMetric(total=2, count=4) => 0.5', str(metric))

  def test_raises_on_non_scalar_value(self):
    with self.assertRaises(TypeError):
      metrics.MeanMetric(total=jnp.array([1]), count=jnp.array([2]))
    with self.assertRaises(TypeError):
      metrics.CountMetric(count=jnp.array([3]))

  def test_masked_count(self):
    count = metrics.masked_count(
        targets=jnp.array([[1, 3], [1, 1], [2, 1], [0, 2]]), mask_values=(0, 2))
    self.assertEqual(count.result(), 5)

  def test_truncation_rate(self):
    targets = jnp.array([[1, 0], [1, 3], [2, 1], [2, 0], [0, 0]])
    eos_value = 3
    pad_value = 0
    truncation_rate = metrics.truncation_rate(targets, eos_value, pad_value)
    self.assertEqual(truncation_rate.result(), 0.75)  # 3 / 4.

  def test_oov_rate(self):
    targets = jnp.array([[1, 0], [1, 3], [3, 1], [0, 2]])
    oov_values = (3,)
    mask_values = (0, 2)
    oov_rate = metrics.oov_rate(targets, oov_values, mask_values)
    self.assertEqual(oov_rate.result(), 0.4)  # 2 / 5.

  def test_sequence_length(self):
    targets = jnp.array([[1, 0], [1, 3], [2, 1], [2, 0], [0, 0]])
    pad_value = 0
    sequence_length = metrics.sequence_length(targets, pad_value)
    self.assertEqual(sequence_length.result(), 1.5)  # 6 / 4.


if __name__ == '__main__':
  tf.test.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Experimental modules for fedjax.

APIs in the experimental package are subject to changes without guarantee of
backward compatibility. Certain modules may not even work.
"""

from fedjax.experimental import client_datasets
from fedjax.experimental import federated_algorithm
from fedjax.experimental import federated_data
from fedjax.experimental import for_each_client
from fedjax.experimental import metrics
from fedjax.experimental import model
from fedjax.experimental import optimizers
from fedjax.experimental import serialization
from fedjax.experimental import sqlite_federated_data
from fedjax.experimental import typing



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.aggregators.aggregator."""
from absl.testing import absltest

from fedjax import core
from fedjax.experimental.aggregators import aggregator


class AggregatorTest(absltest.TestCase):

  def test_aggregator(self):
    num_classes = 10
    num_clients = 10
    num_examples = 5
    data, model = core.test_util.create_toy_example(
        num_clients=10,
        num_clusters=4,
        num_classes=num_classes,
        num_examples=5,
        seed=0)
    rng_seq = core.PRNGSequence(0)
    client_optimizer = core.get_optimizer(
        core.OptimizerName.SGD, learning_rate=0.1)
    client_trainer = core.DefaultClientTrainer(model, client_optimizer)
    init_params = model.init_params(0)
    client_outputs = core.train_multiple_clients(
        federated_data=data,
        client_ids=data.client_ids,
        client_trainer=client_trainer,
        init_client_trainer_state=client_trainer.init_state(init_params),
        rng_seq=rng_seq,
        client_data_hparams=core.ClientDataHParams(batch_size=10))

    def get_delta_params_and_weight(client_output):
      delta_params = core.tree_multimap(lambda a, b: a - b, init_params,
                                        client_output.params)
      return delta_params, client_output.num_examples

    delta_params_and_weight = map(get_delta_params_and_weight, client_outputs)

    default_aggregator = aggregator.MeanAggregator()
    init_aggregator_state = default_aggregator.init_state()
    rng_seq2 = core.PRNGSequence(1)
    _, new_state = default_aggregator.aggregate(init_aggregator_state,
                                                delta_params_and_weight,
                                                rng_seq2)
    self.assertEqual(new_state.total_weight, num_clients * num_examples)


if __name__ == '__main__':
  absltest.main()
